%%Linear Quadratic Regulator
%an LQR is an implementation of control using the state space model. it
%attmepts to move the current state to some desired state. for conveniece
%the desired state is generally set to [0,0....], As maniplating the system
%to consider any other state from here is trivial (current-desired) -> 0
%the system from the kalman tutorial is considered here


clear all;
clc;
clf;

N = 2000;
Ts = 0.002; %500Hz sampling
Time = 1:1:N;
Time = Time*Ts;

A = [0 1 0 0;
    0 0 0 0;
    0 0 0 1;
    0 0 0 0];

B = [0 1 0 0;
    0 0 0 1]'; %we have the ability to apply a force in the X axis and Y axis

C = [1 0 0 0;
    0 0 1 0]; %and are measuring the objects position

D = 0;
sys = ss(A,B,C,D);

sysd = c2d(sys,Ts); %Discretise the continuous ss sys


input = [zeros(5,2);ones(N-5,2)]';


Q = [0 0 0 0;
    0 0.1 0 0;
    0 0 0 0;
    0 0 0 0.1]; %Covariance state

Q = Q + ones(4)*10^-10;
WQ = Q^(-1/2);
DWQ = inv(WQ);

noiseState = DWQ*randn(4,N);

R = [.1 0;
    0 .2]; %covariance sensors
R = R + ones(2)*10^-10;
WR = R^(-1/2); %take a variable with covar mat R to a whitened state
DWR = inv(WR); %dewhiten a white unit power GRV to covar R


noiseSensors = DWR*randn(2,N); %producing noise with the given covariance


%%Steady state KF
Lupanyov = zeros(4);
for i =1:1:1000
    Lupanyov = sysd.A*Lupanyov*sysd.A' + Q - sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R)*sysd.C*Lupanyov*sysd.A';
end
kalmanGain = sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R); %our state estimate filter



%%Steady state LQR gain
Q = zeros(4);
Q(1,1) = 100;
Q(2,2) = 0.2;
Q(3,3) = 100;
Q(4,4) = 0.3;

R = zeros(2);
R(1,1) = 1;
R(2,2) = 2;

LQRGain = lqrd(A,B,Q,R,Ts);


state = [20;0;-20;0]; %True inital state 
estimatedState = [0;0;0;0]; %%inital state estimate

input = [0;0];
%%We cannot simulate the motion independent of the controller+kalman
%%filter, here it will all be considered as one.
for i = 2:1:N %%motion simulation
    
    state(:,i) = sysd.A*state(:,i-1) + sysd.B*input(:,i-1) + noiseState(:,i-1); %compute the true state position
    Measurement(:,i) = sysd.C*state(:,i-1) + noiseSensors(:,i-1);   %compute the observed position of the plant

    predictedState = sysd.A*estimatedState(:,i-1) + sysd.B*input(:,i-1); % predict the system position. note the control effort input is included in the state prediction    
    innovation = Measurement(:,i) - sysd.C*predictedState; 
    estimatedState(:,i) = predictedState + kalmanGain*innovation; %update system position accounting for the measured position
    
    
    input(:,i) = -LQRGain*estimatedState(:,i); %compute the optimal control effort input given the LQRGain. note the LQRgain is fixed, it does not depend on the current state
end

figure(1)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,estimatedState([1,3],:)','b--');

title('Displacement of the system when controlled by an LQR')
xlabel('Time (s)') % x-axis label
ylabel('Displacement(m)') % y-axis label
legend('Measured Disp in X','Measured Disp in Y','True position X','True position Y','Estimated position X','Estimated position Y');



figure(2)
plot(Time,input');

title('Control Effort commanded by the LQR')
xlabel('Time (s)') % x-axis label
ylabel('Effort (N)') % y-axis label
legend('Force applied along X axis','Force Applied along Y axis');



figure(3)
title('Animation of system motion on the X-Y plane')
xlabel('X position (m)') % x-axis label
ylabel('Y position (m)') % y-axis label
grid on
legend('True path','estimated path','Measurements');

a1 = animatedline('Color',[0 0 0]);
a2 = animatedline('Color',[0 .5 .5]);
a3 = animatedline('Color',[1 1 1],'marker','+','MarkerEdgeColor',[1 0 0]);

axis([-30 30 -30 30])
for k = 1:N;
    % true state
    xk = state(1,k);
    yst = state(3,k);
    addpoints(a1,xk,yst);

    % estimated state
    xk = estimatedState(1,k);
    yest = estimatedState(3,k);
    addpoints(a2,xk,yest);
    
    %estimated points
    if(mod(k,2) == 0)
        addpoints(a3,Measurement(1,k),Measurement(2,k));
    end
    
    % update screen
    drawnow %limitrate
    %pause(0.1)
end


